# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import jsonpath_ng
import pytest

from airflow.providers.amazon.aws.utils.sqs import (
    filter_messages,
    filter_messages_jsonpath,
    filter_messages_literal,
    process_response,
)


@pytest.fixture
def response_json():
    return {"Messages": [{"Body": "message1"}, {"Body": "message2"}]}


@pytest.fixture
def response_nested_dict():
    return [{"Body": '{"key": "value1"}'}, {"Body": '{"key": "value2"}'}]


@pytest.fixture
def response_dict():
    return [{"Body": "message1"}, {"Body": "message2"}, {"Body": "message3"}]


def test_process_response_with_empty_response():
    response_json = {}
    processed_response = process_response(response_json)
    assert processed_response == []


def test_process_response_with_no_messages(response_json):
    response_json["Messages"] = []
    processed_response = process_response(response_json)
    assert processed_response == []


def test_process_response_with_messages_and_no_filtering(response_json):
    processed_response = process_response(response_json)
    assert processed_response == response_json["Messages"]


def test_process_response_with_messages_and_literal_filtering(response_json):
    processed_response = process_response(
        response_json, message_filtering="literal", message_filtering_match_values=["message1"]
    )
    assert processed_response == [{"Body": "message1"}]


def test_filter_messages_literal():
    messages = [{"Body": "message1"}, {"Body": "message2"}]
    filtered_messages = filter_messages(
        messages,
        message_filtering="literal",
        message_filtering_match_values=["message1"],
        message_filtering_config="",
    )
    assert filtered_messages == [{"Body": "message1"}]


def test_filter_messages_jsonpath(response_nested_dict):
    filtered_messages = filter_messages(
        response_nested_dict,
        message_filtering="jsonpath",
        message_filtering_match_values=["value1"],
        message_filtering_config="key",
    )
    assert filtered_messages == [{"Body": '{"key": "value1"}'}]


def test_filter_messages_literal_with_matching_values(response_dict):
    filtered_messages = filter_messages_literal(
        response_dict, message_filtering_match_values=["message1", "message3"]
    )
    assert filtered_messages == [{"Body": "message1"}, {"Body": "message3"}]


def test_filter_messages_literal_with_no_matching_values(response_dict):
    filtered_messages = filter_messages_literal(response_dict, message_filtering_match_values=["message4"])
    assert filtered_messages == []


def test_filter_messages_literal_with_empty_messages():
    messages = []
    filtered_messages = filter_messages_literal(messages, message_filtering_match_values=["message1"])
    assert filtered_messages == []


def test_filter_messages_jsonpath_with_matching_values(response_nested_dict):
    filtered_messages = filter_messages_jsonpath(
        response_nested_dict,
        message_filtering_match_values=["value1"],
        message_filtering_config="key",
        parse=jsonpath_ng.parse,
    )
    assert filtered_messages == [{"Body": '{"key": "value1"}'}]


def test_filter_messages_jsonpath_with_no_matching_values(response_nested_dict):
    filtered_messages = filter_messages_jsonpath(
        response_nested_dict,
        message_filtering_match_values=["value3"],
        message_filtering_config="key",
        parse=jsonpath_ng.parse,
    )
    assert filtered_messages == []


def test_filter_messages_jsonpath_with_empty_messages():
    messages = []
    filtered_messages = filter_messages_jsonpath(
        messages,
        message_filtering_match_values=["value1"],
        message_filtering_config="key",
        parse=jsonpath_ng.parse,
    )
    assert filtered_messages == []



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator
from functools import cached_property
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.triggers.base import BaseTrigger, TriggerEvent

if TYPE_CHECKING:
    from datetime import datetime


class S3KeyTrigger(BaseTrigger):
    """
    S3KeyTrigger is fired as deferred class with params to run the task in trigger worker.

    :param bucket_name: Name of the S3 bucket. Only needed when ``bucket_key``
        is not provided as a full s3:// url.
    :param bucket_key:  The key being waited on. Supports full s3:// style url
        or relative path from root level. When it's specified as a full s3://
        url, please leave bucket_name as `None`.
    :param wildcard_match: whether the bucket_key should be interpreted as a
        Unix wildcard pattern
    :param aws_conn_id: reference to the s3 connection
    :param use_regex: whether to use regex to check bucket
    :param metadata_keys: List of head_object attributes to gather and send to ``check_fn``.
        Acceptable values: Any top level attribute returned by s3.head_object. Specify * to return
        all available attributes.
        Default value: "Size".
        If the requested attribute is not found, the key is still included and the value is None.
    :param hook_params: params for hook its optional
    """

    def __init__(
        self,
        bucket_name: str,
        bucket_key: str | list[str],
        wildcard_match: bool = False,
        aws_conn_id: str | None = "aws_default",
        poke_interval: float = 5.0,
        should_check_fn: bool = False,
        use_regex: bool = False,
        region_name: str | None = None,
        verify: bool | str | None = None,
        botocore_config: dict | None = None,
        metadata_keys: list[str] | None = None,
        **hook_params: Any,
    ):
        super().__init__()
        self.bucket_name = bucket_name
        self.bucket_key = bucket_key
        self.wildcard_match = wildcard_match
        self.aws_conn_id = aws_conn_id
        self.hook_params = hook_params
        self.poke_interval = poke_interval
        self.should_check_fn = should_check_fn
        self.use_regex = use_regex
        self.region_name = region_name
        self.verify = verify
        self.botocore_config = botocore_config
        self.metadata_keys = metadata_keys if metadata_keys else ["Size", "Key"]

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize S3KeyTrigger arguments and classpath."""
        return (
            "airflow.providers.amazon.aws.triggers.s3.S3KeyTrigger",
            {
                "bucket_name": self.bucket_name,
                "bucket_key": self.bucket_key,
                "wildcard_match": self.wildcard_match,
                "aws_conn_id": self.aws_conn_id,
                "hook_params": self.hook_params,
                "poke_interval": self.poke_interval,
                "should_check_fn": self.should_check_fn,
                "use_regex": self.use_regex,
                "region_name": self.region_name,
                "verify": self.verify,
                "botocore_config": self.botocore_config,
                "metadata_keys": self.metadata_keys,
            },
        )

    @cached_property
    def hook(self) -> S3Hook:
        return S3Hook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Make an asynchronous connection using S3HookAsync."""
        try:
            async with await self.hook.get_async_conn() as client:
                while True:
                    if await self.hook.check_key_async(
                        client, self.bucket_name, self.bucket_key, self.wildcard_match, self.use_regex
                    ):
                        if self.should_check_fn:
                            raw_objects = await self.hook.get_files_async(
                                client, self.bucket_name, self.bucket_key, self.wildcard_match
                            )
                            files = []
                            for f in raw_objects:
                                metadata = {}
                                obj = await self.hook.get_head_object_async(
                                    client=client, key=f, bucket_name=self.bucket_name
                                )
                                if obj is None:
                                    return

                                if "*" in self.metadata_keys:
                                    metadata = obj
                                else:
                                    for mk in self.metadata_keys:
                                        if mk == "Size":
                                            metadata[mk] = obj.get("ContentLength")
                                        else:
                                            metadata[mk] = obj.get(mk, None)
                                metadata["Key"] = f
                                files.append(metadata)
                            await asyncio.sleep(self.poke_interval)
                            yield TriggerEvent({"status": "running", "files": files})
                        else:
                            yield TriggerEvent({"status": "success"})
                        return

                    self.log.info("Sleeping for %s seconds", self.poke_interval)
                    await asyncio.sleep(self.poke_interval)
        except Exception as e:
            yield TriggerEvent({"status": "error", "message": str(e)})


class S3KeysUnchangedTrigger(BaseTrigger):
    """
    S3KeysUnchangedTrigger is fired as deferred class with params to run the task in trigger worker.

    :param bucket_name: Name of the S3 bucket. Only needed when ``bucket_key``
        is not provided as a full s3:// url.
    :param prefix: The prefix being waited on. Relative path from bucket root level.
    :param inactivity_period: The total seconds of inactivity to designate
        keys unchanged. Note, this mechanism is not real time and
        this operator may not return until a poke_interval after this period
        has passed with no additional objects sensed.
    :param min_objects: The minimum number of objects needed for keys unchanged
        sensor to be considered valid.
    :param inactivity_seconds: reference to the seconds of inactivity
    :param previous_objects: The set of object ids found during the last poke.
    :param allow_delete: Should this sensor consider objects being deleted
    :param aws_conn_id: reference to the s3 connection
    :param last_activity_time: last modified or last active time
    :param verify: Whether or not to verify SSL certificates for S3 connection.
        By default SSL certificates are verified.
    :param hook_params: params for hook its optional
    """

    def __init__(
        self,
        bucket_name: str,
        prefix: str,
        inactivity_period: float = 60 * 60,
        min_objects: int = 1,
        inactivity_seconds: int = 0,
        previous_objects: set[str] | None = None,
        allow_delete: bool = True,
        aws_conn_id: str | None = "aws_default",
        last_activity_time: datetime | None = None,
        region_name: str | None = None,
        verify: bool | str | None = None,
        botocore_config: dict | None = None,
        **hook_params: Any,
    ):
        super().__init__()
        self.bucket_name = bucket_name
        self.prefix = prefix
        if inactivity_period < 0:
            raise ValueError("inactivity_period must be non-negative")
        if not previous_objects:
            previous_objects = set()
        self.inactivity_period = inactivity_period
        self.min_objects = min_objects
        self.previous_objects = previous_objects
        self.inactivity_seconds = inactivity_seconds
        self.allow_delete = allow_delete
        self.aws_conn_id = aws_conn_id
        self.last_activity_time = last_activity_time
        self.polling_period_seconds = 0
        self.region_name = region_name
        self.verify = verify
        self.botocore_config = botocore_config
        self.hook_params = hook_params

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize S3KeysUnchangedTrigger arguments and classpath."""
        return (
            "airflow.providers.amazon.aws.triggers.s3.S3KeysUnchangedTrigger",
            {
                "bucket_name": self.bucket_name,
                "prefix": self.prefix,
                "inactivity_period": self.inactivity_period,
                "min_objects": self.min_objects,
                "previous_objects": self.previous_objects,
                "inactivity_seconds": self.inactivity_seconds,
                "allow_delete": self.allow_delete,
                "aws_conn_id": self.aws_conn_id,
                "last_activity_time": self.last_activity_time,
                "hook_params": self.hook_params,
                "polling_period_seconds": self.polling_period_seconds,
                "region_name": self.region_name,
                "verify": self.verify,
                "botocore_config": self.botocore_config,
            },
        )

    @cached_property
    def hook(self) -> S3Hook:
        return S3Hook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Make an asynchronous connection using S3Hook."""
        try:
            async with await self.hook.get_async_conn() as client:
                while True:
                    result = await self.hook.is_keys_unchanged_async(
                        client=client,
                        bucket_name=self.bucket_name,
                        prefix=self.prefix,
                        inactivity_period=self.inactivity_period,
                        min_objects=self.min_objects,
                        previous_objects=self.previous_objects,
                        inactivity_seconds=self.inactivity_seconds,
                        allow_delete=self.allow_delete,
                        last_activity_time=self.last_activity_time,
                    )
                    if result.get("status") in ("success", "error"):
                        yield TriggerEvent(result)
                        return
                    elif result.get("status") == "pending":
                        self.previous_objects = result.get("previous_objects", set())
                        self.last_activity_time = result.get("last_activity_time")
                        self.inactivity_seconds = result.get("inactivity_seconds", 0)
                    await asyncio.sleep(self.polling_period_seconds)
        except Exception as e:
            yield TriggerEvent({"status": "error", "message": str(e)})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import time
from collections.abc import Generator
from datetime import datetime, timedelta, timezone
from threading import Event, Thread
from typing import TYPE_CHECKING

from botocore.exceptions import ClientError, ConnectionClosedError

from airflow.providers.amazon.aws.hooks.logs import AwsLogsHook

if TYPE_CHECKING:
    from airflow.sdk.types import Logger


class AwsTaskLogFetcher(Thread):
    """Fetch Cloudwatch log events with specific interval and send the log events to the logger.info."""

    def __init__(
        self,
        *,
        log_group: str,
        log_stream_name: str,
        fetch_interval: timedelta,
        logger: Logger,
        aws_conn_id: str | None = "aws_default",
        region_name: str | None = None,
    ):
        super().__init__()
        self._event = Event()

        self.fetch_interval = fetch_interval

        self.logger = logger
        self.log_group = log_group
        self.log_stream_name = log_stream_name

        self.hook = AwsLogsHook(aws_conn_id=aws_conn_id, region_name=region_name)

    def run(self) -> None:
        continuation_token = AwsLogsHook.ContinuationToken()
        while not self.is_stopped():
            time.sleep(self.fetch_interval.total_seconds())
            log_events = self._get_log_events(continuation_token)
            prev_timestamp_event = None
            for log_event in log_events:
                current_timestamp_event = datetime.fromtimestamp(
                    log_event["timestamp"] / 1000.0, tz=timezone.utc
                )
                if current_timestamp_event == prev_timestamp_event:
                    # When multiple events have the same timestamp, somehow, only one event is logged
                    # As a consequence, some logs are missed in the log group (in case they have the same
                    # timestamp)
                    # When a slight delay is added before logging the event, that solves the issue
                    # See https://github.com/apache/airflow/issues/40875
                    time.sleep(0.001)
                self.logger.info(self.event_to_str(log_event))
                prev_timestamp_event = current_timestamp_event

    def _get_log_events(self, skip_token: AwsLogsHook.ContinuationToken | None = None) -> Generator:
        if skip_token is None:
            skip_token = AwsLogsHook.ContinuationToken()
        try:
            yield from self.hook.get_log_events(
                self.log_group, self.log_stream_name, continuation_token=skip_token
            )
        except ClientError as error:
            if error.response["Error"]["Code"] != "ResourceNotFoundException":
                self.logger.warning("Error on retrieving Cloudwatch log events", error)
            else:
                self.logger.info(
                    "Cannot find log stream yet, it can take a couple of seconds to show up. "
                    "If this error persists, check that the log group and stream are correct: "
                    "group: %s\tstream: %s",
                    self.log_group,
                    self.log_stream_name,
                )
            yield from ()
        except ConnectionClosedError as error:
            self.logger.warning("ConnectionClosedError on retrieving Cloudwatch log events", error)
            yield from ()

    @staticmethod
    def event_to_str(event: dict) -> str:
        event_dt = datetime.fromtimestamp(event["timestamp"] / 1000.0, tz=timezone.utc)
        formatted_event_dt = event_dt.strftime("%Y-%m-%d %H:%M:%S,%f")[:-3]
        message = event["message"]
        return f"[{formatted_event_dt}] {message}"

    def get_last_log_messages(self, number_messages) -> list:
        """
        Get the last logs messages in one single request.

         NOTE: some restrictions apply:
         - if logs are too old, the response will be empty
         - the max number of messages we can retrieve is constrained by cloudwatch limits (10,000).
        """
        response = self.hook.conn.get_log_events(
            logGroupName=self.log_group,
            logStreamName=self.log_stream_name,
            startFromHead=False,
            limit=number_messages,
        )
        return [log["message"] for log in response["events"]]

    def get_last_log_message(self) -> str | None:
        try:
            return self.get_last_log_messages(1)[0]
        except IndexError:
            return None

    def is_stopped(self) -> bool:
        return self._event.is_set()

    def stop(self):
        self._event.set()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import ANY

import pytest

from airflow.providers.amazon.aws.hooks.glue import GlueJobHook
from airflow.providers.amazon.aws.sensors.glue import GlueJobSensor
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred


class TestGlueJobSensor:
    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_poke(self, mock_get_job_state, mock_conn, mock_print_job_logs):
        mock_conn.return_value.get_job_run()
        mock_get_job_state.return_value = "SUCCEEDED"
        op = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name="aws_test_glue_job",
            run_id="5152fgsfsjhsh61661",
            poke_interval=1,
            timeout=5,
        )

        assert op.poke({})
        mock_print_job_logs.assert_not_called()

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_poke_with_verbose_logging(self, mock_get_job_state, mock_conn, mock_print_job_logs):
        mock_conn.return_value.get_job_run()
        mock_get_job_state.return_value = "SUCCEEDED"
        job_name = "job_name"
        job_run_id = "job_run_id"
        op = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            poke_interval=1,
            timeout=5,
            verbose=True,
        )

        assert op.poke({})
        mock_print_job_logs.assert_called_once_with(
            job_name=job_name,
            run_id=job_run_id,
            continuation_tokens=ANY,
        )

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_poke_false(self, mock_get_job_state, mock_conn, mock_print_job_logs):
        mock_conn.return_value.get_job_run()
        mock_get_job_state.return_value = "RUNNING"
        op = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name="aws_test_glue_job",
            run_id="5152fgsfsjhsh61661",
            poke_interval=1,
            timeout=5,
        )

        assert not op.poke({})
        mock_print_job_logs.assert_not_called()

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_poke_false_with_verbose_logging(self, mock_get_job_state, mock_conn, mock_print_job_logs):
        mock_conn.return_value.get_job_run()
        mock_get_job_state.return_value = "RUNNING"
        job_name = "job_name"
        job_run_id = "job_run_id"
        op = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            poke_interval=1,
            timeout=5,
            verbose=True,
        )

        assert not op.poke({})
        mock_print_job_logs.assert_called_once_with(
            job_name=job_name,
            run_id=job_run_id,
            continuation_tokens=ANY,
        )

    @mock.patch.object(GlueJobHook, "print_job_logs")
    @mock.patch.object(GlueJobHook, "get_conn")
    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_poke_failed_job_with_verbose_logging(self, mock_get_job_state, mock_conn, mock_print_job_logs):
        mock_conn.return_value.get_job_run()
        mock_get_job_state.return_value = "FAILED"
        job_name = "job_name"
        job_run_id = "job_run_id"
        op = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            poke_interval=1,
            timeout=5,
            verbose=True,
        )

        with pytest.raises(AirflowException):
            assert not op.poke({})
        mock_print_job_logs.assert_called_once_with(
            job_name=job_name,
            run_id=job_run_id,
            continuation_tokens=ANY,
        )

    @mock.patch("airflow.providers.amazon.aws.hooks.glue.GlueJobHook.get_job_state")
    def test_fail_poke(self, get_job_state):
        job_name = "job_name"
        job_run_id = "job_run_id"
        op = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            poke_interval=1,
            timeout=5,
            verbose=True,
        )
        op.verbose = False
        job_state = "FAILED"
        get_job_state.return_value = job_state
        job_error_message = "Exiting Job"
        with pytest.raises(AirflowException, match=job_error_message):
            op.poke(context={})

    def test_deferrable_execute_raises_task_deferred(self):
        job_name = "job_name"
        job_run_id = "job_run_id"
        sensor = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            deferrable=True,
            poke_interval=1,
            timeout=5,
        )
        with pytest.raises(TaskDeferred):
            sensor.execute({})

    @mock.patch.object(GlueJobSensor, "defer")
    def test_default_timeout(self, mock_defer):
        mock_defer.side_effect = TaskDeferred(trigger=mock.Mock(), method_name="execute_complete")
        sensor = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name="job_name",
            run_id="job_run_id",
            deferrable=True,
            poke_interval=5,
            max_retries=30,
        )
        with pytest.raises(TaskDeferred):
            sensor.execute({})
        call_kwargs = mock_defer.call_args.kwargs["trigger"]
        assert call_kwargs.attempts == 30
        mock_defer.assert_called_once()

    def test_default_args(self):
        job_name = "job_name"
        job_run_id = "job_run_id"
        sensor = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
        )
        assert sensor.poke_interval == 120
        assert sensor.verbose is False
        assert sensor.deferrable is False or isinstance(sensor.deferrable, bool)
        assert sensor.aws_conn_id == "aws_default"

    def test_custom_args(self):
        job_name = "job_name"
        job_run_id = "job_run_id"
        sensor = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            verbose=True,
            deferrable=True,
            poke_interval=10,
            aws_conn_id="custom_conn",
            max_retries=20,
        )
        assert sensor.verbose is True
        assert sensor.deferrable is True
        assert sensor.poke_interval == 10
        assert sensor.aws_conn_id == "custom_conn"
        assert sensor.max_retries == 20

    def test_defferable_params_passed_to_trigger(self):
        job_name = "job_name"
        job_run_id = "job_run_id"
        sensor = GlueJobSensor(
            task_id="test_glue_job_sensor",
            job_name=job_name,
            run_id=job_run_id,
            verbose=True,
            deferrable=True,
            poke_interval=10,
            region_name="us-west-2",
            aws_conn_id="custom_conn",
            max_retries=20,
        )
        with pytest.raises(TaskDeferred) as defer:
            sensor.execute({})

        assert defer.value.trigger.job_name == job_name
        assert defer.value.trigger.run_id == job_run_id
        assert defer.value.trigger.region_name == "us-west-2"
        assert defer.value.trigger.verbose
        assert defer.value.trigger.waiter_delay == 10
        assert defer.value.trigger.attempts == 20
        assert defer.value.trigger.aws_conn_id == "custom_conn"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.sagemaker import SageMakerHook
from airflow.providers.amazon.aws.operators import sagemaker
from airflow.providers.amazon.aws.operators.sagemaker import (
    SageMakerBaseOperator,
    SageMakerProcessingOperator,
)
from airflow.providers.amazon.aws.triggers.sagemaker import SageMakerTrigger
from airflow.providers.common.compat.openlineage.facet import Dataset
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred
from airflow.providers.openlineage.extractors import OperatorLineage

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

CREATE_PROCESSING_PARAMS: dict = {
    "AppSpecification": {
        "ContainerArguments": ["container_arg"],
        "ContainerEntrypoint": ["container_entrypoint"],
        "ImageUri": "image_uri",
    },
    "Environment": {"key": "value"},
    "ExperimentConfig": {
        "ExperimentName": "experiment_name",
        "TrialComponentDisplayName": "trial_component_display_name",
        "TrialName": "trial_name",
    },
    "ProcessingInputs": [
        {
            "InputName": "analytics_input_name",
            "S3Input": {
                "LocalPath": "local_path",
                "S3CompressionType": "None",
                "S3DataDistributionType": "FullyReplicated",
                "S3DataType": "S3Prefix",
                "S3InputMode": "File",
                "S3Uri": "s3_uri",
            },
        }
    ],
    "ProcessingJobName": "job_name",
    "ProcessingOutputConfig": {
        "KmsKeyId": "kms_key_ID",
        "Outputs": [
            {
                "OutputName": "analytics_output_name",
                "S3Output": {
                    "LocalPath": "local_path",
                    "S3UploadMode": "EndOfJob",
                    "S3Uri": "s3_uri",
                },
            }
        ],
    },
    "ProcessingResources": {
        "ClusterConfig": {
            "InstanceCount": "2",
            "InstanceType": "ml.p3.2xlarge",
            "VolumeSizeInGB": "30",
            "VolumeKmsKeyId": "kms_key",
        }
    },
    "RoleArn": "arn:aws:iam::0122345678910:role/SageMakerPowerUser",
    "Tags": [{"key": "value"}],
}

CREATE_PROCESSING_PARAMS_WITH_STOPPING_CONDITION: dict = CREATE_PROCESSING_PARAMS.copy()
CREATE_PROCESSING_PARAMS_WITH_STOPPING_CONDITION.update(StoppingCondition={"MaxRuntimeInSeconds": "3600"})

EXPECTED_INTEGER_FIELDS: list[list[str]] = [
    ["ProcessingResources", "ClusterConfig", "InstanceCount"],
    ["ProcessingResources", "ClusterConfig", "VolumeSizeInGB"],
]
EXPECTED_STOPPING_CONDITION_INTEGER_FIELDS: list[list[str]] = [["StoppingCondition", "MaxRuntimeInSeconds"]]


class TestSageMakerProcessingOperator:
    def setup_method(self):
        self.processing_config_kwargs = dict(
            task_id="test_sagemaker_operator",
            wait_for_completion=False,
            check_interval=5,
        )

        self.defer_processing_config_kwargs = dict(
            task_id="test_sagemaker_operator", wait_for_completion=True, check_interval=5, deferrable=True
        )

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=0)
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_integer_fields_without_stopping_condition(self, _, __, ___, mock_desc):
        mock_desc.side_effect = [ClientError({"Error": {"Code": "ValidationException"}}, "op"), None]
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS
        )

        sagemaker.execute(None)

        assert sagemaker.integer_fields == EXPECTED_INTEGER_FIELDS
        for key1, key2, key3 in EXPECTED_INTEGER_FIELDS:
            assert sagemaker.config[key1][key2][key3] == int(sagemaker.config[key1][key2][key3])

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=0)
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_integer_fields_with_stopping_condition(self, _, __, ___, mock_desc):
        mock_desc.side_effect = [ClientError({"Error": {"Code": "ValidationException"}}, "op"), None]
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS_WITH_STOPPING_CONDITION
        )
        sagemaker.execute(None)
        assert (
            sagemaker.integer_fields == EXPECTED_INTEGER_FIELDS + EXPECTED_STOPPING_CONDITION_INTEGER_FIELDS
        )
        for key1, key2, *key3_raw in EXPECTED_INTEGER_FIELDS:
            if key3_raw:
                (key3,) = key3_raw
                assert sagemaker.config[key1][key2][key3] == int(sagemaker.config[key1][key2][key3])
            else:
                sagemaker.config[key1][key2] == int(sagemaker.config[key1][key2])

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=0)
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_execute(self, _, mock_processing, __, mock_desc):
        mock_desc.side_effect = [ClientError({"Error": {"Code": "ValidationException"}}, "op"), None]
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS
        )
        sagemaker.execute(None)
        mock_processing.assert_called_once_with(
            CREATE_PROCESSING_PARAMS, wait_for_completion=False, check_interval=5, max_ingestion_time=None
        )

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=0)
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_execute_with_stopping_condition(self, _, mock_processing, __, mock_desc):
        mock_desc.side_effect = [ClientError({"Error": {"Code": "ValidationException"}}, "op"), None]
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS_WITH_STOPPING_CONDITION
        )
        sagemaker.execute(None)
        mock_processing.assert_called_once_with(
            CREATE_PROCESSING_PARAMS_WITH_STOPPING_CONDITION,
            wait_for_completion=False,
            check_interval=5,
            max_ingestion_time=None,
        )

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 404}},
    )
    def test_execute_with_failure(self, _, mock_desc):
        mock_desc.side_effect = [ClientError({"Error": {"Code": "ValidationException"}}, "op"), None]
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS
        )
        with pytest.raises(AirflowException):
            sagemaker.execute(None)

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=1)
    @mock.patch.object(
        SageMakerHook, "create_processing_job", return_value={"ResponseMetadata": {"HTTPStatusCode": 200}}
    )
    def test_execute_with_existing_job_timestamp(self, mock_create_processing_job, _, mock_desc):
        mock_desc.side_effect = [None, ClientError({"Error": {"Code": "ValidationException"}}, "op"), None]
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS
        )
        sagemaker.action_if_job_exists = "timestamp"
        sagemaker.execute(None)

        expected_config = CREATE_PROCESSING_PARAMS.copy()
        # Expect to see ProcessingJobName suffixed because we return one existing job
        expected_config["ProcessingJobName"].startswith("job_name-")
        mock_create_processing_job.assert_called_once_with(
            expected_config,
            wait_for_completion=False,
            check_interval=5,
            max_ingestion_time=None,
        )

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=1)
    @mock.patch.object(
        SageMakerHook, "create_processing_job", return_value={"ResponseMetadata": {"HTTPStatusCode": 200}}
    )
    def test_execute_with_existing_job_fail(self, _, __, ___):
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs, config=CREATE_PROCESSING_PARAMS
        )
        sagemaker.action_if_job_exists = "fail"
        with pytest.raises(AirflowException):
            sagemaker.execute(None)

    @mock.patch.object(SageMakerHook, "describe_processing_job")
    def test_action_if_job_exists_validation(self, mock_client):
        with pytest.raises(AirflowException):
            SageMakerProcessingOperator(
                **self.processing_config_kwargs,
                config=CREATE_PROCESSING_PARAMS,
                action_if_job_exists="not_fail_or_increment",
            )

    @mock.patch.object(
        SageMakerHook, "describe_processing_job", return_value={"ProcessingJobStatus": "InProgress"}
    )
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={
            "ProcessingJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        },
    )
    @mock.patch.object(SageMakerBaseOperator, "_check_if_job_exists", return_value=False)
    def test_operator_defer(self, mock_job_exists, mock_processing, mock_describe):
        sagemaker_operator = SageMakerProcessingOperator(
            **self.defer_processing_config_kwargs,
            config=CREATE_PROCESSING_PARAMS,
        )
        sagemaker_operator.wait_for_completion = True
        with pytest.raises(TaskDeferred) as exc:
            sagemaker_operator.execute(context=None)
        assert isinstance(exc.value.trigger, SageMakerTrigger), "Trigger is not a SagemakerTrigger"

    @mock.patch("airflow.providers.amazon.aws.operators.sagemaker.SageMakerProcessingOperator.defer")
    @mock.patch.object(
        SageMakerHook, "describe_processing_job", return_value={"ProcessingJobStatus": "Completed"}
    )
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(SageMakerBaseOperator, "_check_if_job_exists", return_value=False)
    def test_operator_complete_before_defer(
        self, mock_job_exists, mock_processing, mock_describe, mock_defer
    ):
        sagemaker_operator = SageMakerProcessingOperator(
            **self.defer_processing_config_kwargs,
            config=CREATE_PROCESSING_PARAMS,
        )
        sagemaker_operator.execute(context=None)
        assert not mock_defer.called

    @mock.patch("airflow.providers.amazon.aws.operators.sagemaker.SageMakerProcessingOperator.defer")
    @mock.patch.object(
        SageMakerHook,
        "describe_processing_job",
        return_value={"ProcessingJobStatus": "Failed", "FailureReason": "It failed"},
    )
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(SageMakerBaseOperator, "_check_if_job_exists", return_value=False)
    def test_operator_failed_before_defer(
        self,
        mock_job_exists,
        mock_processing,
        mock_describe,
        mock_defer,
    ):
        sagemaker_operator = SageMakerProcessingOperator(
            **self.defer_processing_config_kwargs,
            config=CREATE_PROCESSING_PARAMS,
        )
        with pytest.raises(AirflowException):
            sagemaker_operator.execute(context=None)

        assert not mock_defer.called

    @mock.patch.object(
        SageMakerHook,
        "describe_processing_job",
        return_value={
            "ProcessingInputs": [{"S3Input": {"S3Uri": "s3://input-bucket/input-path"}}],
            "ProcessingOutputConfig": {
                "Outputs": [{"S3Output": {"S3Uri": "s3://output-bucket/output-path"}}]
            },
        },
    )
    @mock.patch.object(SageMakerHook, "count_processing_jobs_by_name", return_value=0)
    @mock.patch.object(
        SageMakerHook,
        "create_processing_job",
        return_value={"ProcessingJobArn": "test_arn", "ResponseMetadata": {"HTTPStatusCode": 200}},
    )
    @mock.patch.object(SageMakerBaseOperator, "_check_if_job_exists", return_value=False)
    def test_operator_openlineage_data(self, check_job_exists, mock_processing, _, mock_desc):
        sagemaker = SageMakerProcessingOperator(
            **self.processing_config_kwargs,
            config=CREATE_PROCESSING_PARAMS,
            deferrable=True,
        )
        sagemaker.execute(context=None)
        assert sagemaker.get_openlineage_facets_on_complete(None) == OperatorLineage(
            inputs=[Dataset(namespace="s3://input-bucket", name="input-path")],
            outputs=[Dataset(namespace="s3://output-bucket", name="output-path")],
        )

    def test_template_fields(self):
        operator = SageMakerProcessingOperator(
            **self.processing_config_kwargs,
            config=CREATE_PROCESSING_PARAMS,
        )
        validate_template_fields(operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.glue_databrew import GlueDataBrewHook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.glue_databrew import GlueDataBrewJobCompleteTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class GlueDataBrewStartJobOperator(AwsBaseOperator[GlueDataBrewHook]):
    """
    Start an AWS Glue DataBrew job.

    AWS Glue DataBrew is a visual data preparation tool that makes it easier
    for data analysts and data scientists to clean and normalize data
    to prepare it for analytics and machine learning (ML).

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GlueDataBrewStartJobOperator`

    :param job_name: unique job name per AWS Account
    :param wait_for_completion: Whether to wait for job run completion. (default: True)
    :param deferrable: If True, the operator will wait asynchronously for the job to complete.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param waiter_delay: Time in seconds to wait between status checks. Default is 30.
    :param waiter_max_attempts: Maximum number of attempts to check for job completion. (default: 60)
    :return: dictionary with key run_id and value of the resulting job's run_id.

    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueDataBrewHook

    template_fields: Sequence[str] = aws_template_fields(
        "job_name",
        "wait_for_completion",
        "waiter_delay",
        "waiter_max_attempts",
        "deferrable",
    )

    def __init__(
        self,
        job_name: str,
        wait_for_completion: bool = True,
        delay: int | None = None,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.job_name = job_name
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable

    def execute(self, context: Context):
        job = self.hook.conn.start_job_run(Name=self.job_name)
        run_id = job["RunId"]

        self.log.info("AWS Glue DataBrew Job: %s. Run Id: %s submitted.", self.job_name, run_id)

        if self.deferrable:
            self.log.info("Deferring job %s with run_id %s", self.job_name, run_id)
            self.defer(
                trigger=GlueDataBrewJobCompleteTrigger(
                    job_name=self.job_name,
                    run_id=run_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )

        elif self.wait_for_completion:
            self.log.info(
                "Waiting for AWS Glue DataBrew Job: %s. Run Id: %s to complete.", self.job_name, run_id
            )
            status = self.hook.job_completion(
                job_name=self.job_name,
                delay=self.waiter_delay,
                run_id=run_id,
                max_attempts=self.waiter_max_attempts,
            )
            self.log.info("Glue DataBrew Job: %s status: %s", self.job_name, status)

        return {"run_id": run_id}

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> dict[str, str]:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException("Error while running AWS Glue DataBrew job: %s", validated_event)

        run_id = validated_event.get("run_id", "")
        status = validated_event.get("status", "")

        self.log.info("AWS Glue DataBrew runID: %s completed with status: %s", run_id, status)

        return {"run_id": run_id}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains AWS Firehose hook."""

from __future__ import annotations

import warnings

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.amazon.aws.hooks.firehose import FirehoseHook as _FirehoseHook


class FirehoseHook(_FirehoseHook):
    """
    Interact with Amazon Kinesis Firehose.

    Provide thick wrapper around :external+boto3:py:class:`boto3.client("firehose") <Firehose.Client>`.

    :param delivery_stream: Name of the delivery stream

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    .. deprecated::
        This hook was moved. Import from
       :class:`airflow.providers.amazon.aws.hooks.firehose.FirehoseHook`
       instead of kinesis.py
    """

    def __init__(self, *args, **kwargs) -> None:
        warnings.warn(
            "Importing FirehoseHook from kinesis.py is deprecated "
            "and will be removed in a future release. "
            "Please import it from firehose.py instead.",
            AirflowProviderDeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


class KinesisHook(AwsBaseHook):
    """
    Interact with Amazon Kinesis.

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("kinesis") <Kinesis.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs["client_type"] = "kinesis"
        super().__init__(*args, **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import subprocess
from datetime import datetime

import boto3

from airflow.providers.amazon.aws.hooks.eks import ClusterStates, NodegroupStates
from airflow.providers.amazon.aws.operators.eks import EksCreateClusterOperator, EksDeleteClusterOperator
from airflow.providers.amazon.aws.operators.emr import EmrContainerOperator, EmrEksCreateClusterOperator
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.eks import EksClusterStateSensor, EksNodegroupStateSensor
from airflow.providers.amazon.aws.sensors.emr import EmrContainerSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2.10 compat
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]
try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_emr_eks"

# Externally fetched variables
ROLE_ARN_KEY = "ROLE_ARN"
JOB_ROLE_ARN_KEY = "JOB_ROLE_ARN"
JOB_ROLE_NAME_KEY = "JOB_ROLE_NAME"
SUBNETS_KEY = "SUBNETS"

sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(ROLE_ARN_KEY)
    .add_variable(JOB_ROLE_ARN_KEY)
    .add_variable(JOB_ROLE_NAME_KEY)
    .add_variable(SUBNETS_KEY, split_string=True)
    .build()
)

S3_FILE_NAME = "pi.py"
S3_FILE_CONTENT = """
k = 1
s = 0

for i in range(1000000):
    if i % 2 == 0:
        s += 4/k
    else:
        s -= 4/k

    k += 2

print(s)
"""


@task
def create_launch_template(template_name: str):
    # This launch template enables IMDSv2.
    boto3.client("ec2").create_launch_template(
        LaunchTemplateName=template_name,
        LaunchTemplateData={
            "MetadataOptions": {"HttpEndpoint": "enabled", "HttpTokens": "required"},
        },
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_launch_template(template_name: str):
    boto3.client("ec2").delete_launch_template(LaunchTemplateName=template_name)


@task
def run_eksctl_commands(cluster_name, ns):
    # Install eksctl and enable access for EMR on EKS
    # See https://docs.aws.amazon.com/emr/latest/EMR-on-EKS-DevelopmentGuide/setting-up-cluster-access.html
    file = "https://github.com/weaveworks/eksctl/releases/latest/download/eksctl_$(uname -s)_amd64.tar.gz"
    commands = f"""
        curl --silent --location "{file}" | tar xz -C /tmp &&
        /tmp/eksctl create iamidentitymapping --cluster {cluster_name} --namespace {ns} --service-name "emr-containers" &&
        /tmp/eksctl utils associate-iam-oidc-provider --cluster {cluster_name} --approve
    """

    build = subprocess.Popen(
        commands,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _, err = build.communicate()

    if build.returncode != 0:
        raise RuntimeError(err)


@task
def delete_iam_oidc_identity_provider(cluster_name):
    oidc_provider_issuer_url = boto3.client("eks").describe_cluster(
        name=cluster_name,
    )["cluster"]["identity"]["oidc"]["issuer"]
    oidc_provider_issuer_endpoint = oidc_provider_issuer_url.replace("https://", "")

    account_id = boto3.client("sts").get_caller_identity()["Account"]
    boto3.client("iam").delete_open_id_connect_provider(
        OpenIDConnectProviderArn=f"arn:aws:iam::{account_id}:oidc-provider/{oidc_provider_issuer_endpoint}"
    )


@task
def update_trust_policy_execution_role(cluster_name, cluster_namespace, role_name):
    # Remove any already existing trusted entities added with "update-role-trust-policy"
    # Prevent getting an error "Cannot exceed quota for ACLSizePerRole"
    client = boto3.client("iam")
    role_trust_policy = client.get_role(RoleName=role_name)["Role"]["AssumeRolePolicyDocument"]
    # We assume if the action is sts:AssumeRoleWithWebIdentity, the statement had been added with
    # "update-role-trust-policy". Removing it to not exceed the quota
    role_trust_policy["Statement"] = [
        statement
        for statement in role_trust_policy["Statement"]
        if statement["Action"] != "sts:AssumeRoleWithWebIdentity"
    ]

    client.update_assume_role_policy(
        RoleName=role_name,
        PolicyDocument=json.dumps(role_trust_policy),
    )

    # See https://docs.aws.amazon.com/emr/latest/EMR-on-EKS-DevelopmentGuide/setting-up-trust-policy.html
    # The action "update-role-trust-policy" is not available in boto3, thus we need to do it using AWS CLI
    commands = (
        f"aws emr-containers update-role-trust-policy --cluster-name {cluster_name} "
        f"--namespace {cluster_namespace} --role-name {role_name}"
    )

    build = subprocess.Popen(
        commands,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _, err = build.communicate()

    if build.returncode != 0:
        raise RuntimeError(err)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_virtual_cluster(virtual_cluster_id):
    boto3.client("emr-containers").delete_virtual_cluster(
        id=virtual_cluster_id,
    )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    role_arn = test_context[ROLE_ARN_KEY]
    subnets = test_context[SUBNETS_KEY]
    job_role_arn = test_context[JOB_ROLE_ARN_KEY]
    job_role_name = test_context[JOB_ROLE_NAME_KEY]

    s3_bucket_name = f"{env_id}-bucket"
    eks_cluster_name = f"{env_id}-cluster"
    virtual_cluster_name = f"{env_id}-virtual-cluster"
    nodegroup_name = f"{env_id}-nodegroup"
    eks_namespace = "default"
    launch_template_name = f"{env_id}-launch-template"

    # [START howto_operator_emr_eks_config]
    job_driver_arg = {
        "sparkSubmitJobDriver": {
            "entryPoint": f"s3://{s3_bucket_name}/{S3_FILE_NAME}",
            "sparkSubmitParameters": "--conf spark.executors.instances=2 --conf spark.executors.memory=2G "
            "--conf spark.executor.cores=2 --conf spark.driver.cores=1",
        }
    }

    configuration_overrides_arg = {
        "monitoringConfiguration": {
            "cloudWatchMonitoringConfiguration": {
                "logGroupName": "/emr-eks-jobs",
                "logStreamNamePrefix": "airflow",
            }
        },
    }
    # [END howto_operator_emr_eks_config]

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=s3_bucket_name,
    )

    upload_s3_file = S3CreateObjectOperator(
        task_id="upload_s3_file",
        s3_bucket=s3_bucket_name,
        s3_key=S3_FILE_NAME,
        data=S3_FILE_CONTENT,
    )

    create_cluster_and_nodegroup = EksCreateClusterOperator(
        task_id="create_cluster_and_nodegroup",
        cluster_name=eks_cluster_name,
        nodegroup_name=nodegroup_name,
        cluster_role_arn=role_arn,
        # Opting to use the same ARN for the cluster and the nodegroup here,
        # but a different ARN could be configured and passed if desired.
        nodegroup_role_arn=role_arn,
        resources_vpc_config={"subnetIds": subnets},
        # The launch template enforces IMDSv2 and is required for internal
        # compliance when running these system tests on AWS infrastructure.
        create_nodegroup_kwargs={"launchTemplate": {"name": launch_template_name}},
    )

    await_create_nodegroup = EksNodegroupStateSensor(
        task_id="await_create_nodegroup",
        cluster_name=eks_cluster_name,
        nodegroup_name=nodegroup_name,
        target_state=NodegroupStates.ACTIVE,
        poke_interval=10,
    )

    # [START howto_operator_emr_eks_create_cluster]
    create_emr_eks_cluster = EmrEksCreateClusterOperator(
        task_id="create_emr_eks_cluster",
        virtual_cluster_name=virtual_cluster_name,
        eks_cluster_name=eks_cluster_name,
        eks_namespace=eks_namespace,
    )
    # [END howto_operator_emr_eks_create_cluster]

    # [START howto_operator_emr_container]
    job_starter = EmrContainerOperator(
        task_id="start_job",
        virtual_cluster_id=str(create_emr_eks_cluster.output),
        execution_role_arn=job_role_arn,
        release_label="emr-7.0.0-latest",
        job_driver=job_driver_arg,
        configuration_overrides=configuration_overrides_arg,
        name="pi.py",
    )
    # [END howto_operator_emr_container]
    job_starter.wait_for_completion = False
    job_starter.job_retry_max_attempts = 5

    # [START howto_sensor_emr_container]
    job_waiter = EmrContainerSensor(
        task_id="job_waiter",
        virtual_cluster_id=str(create_emr_eks_cluster.output),
        job_id=str(job_starter.output),
    )
    # [END howto_sensor_emr_container]

    delete_eks_cluster = EksDeleteClusterOperator(
        task_id="delete_eks_cluster",
        cluster_name=eks_cluster_name,
        force_delete_compute=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    await_delete_eks_cluster = EksClusterStateSensor(
        task_id="await_delete_eks_cluster",
        cluster_name=eks_cluster_name,
        target_state=ClusterStates.NONEXISTENT,
        trigger_rule=TriggerRule.ALL_DONE,
        poke_interval=10,
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=s3_bucket_name,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        upload_s3_file,
        create_launch_template(launch_template_name),
        create_cluster_and_nodegroup,
        await_create_nodegroup,
        run_eksctl_commands(eks_cluster_name, eks_namespace),
        update_trust_policy_execution_role(eks_cluster_name, eks_namespace, job_role_name),
        # TEST BODY
        create_emr_eks_cluster,
        job_starter,
        job_waiter,
        # TEST TEARDOWN
        delete_iam_oidc_identity_provider(eks_cluster_name),
        delete_virtual_cluster(str(create_emr_eks_cluster.output)),
        delete_eks_cluster,
        await_delete_eks_cluster,
        delete_launch_template(launch_template_name),
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.ssm import SsmHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.ssm import SsmRunCommandTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class SsmRunCommandCompletedSensor(AwsBaseSensor[SsmHook]):
    """
    Poll the state of an AWS SSM Run Command until completion.

    Waits until all instance jobs reach a terminal state. Fails if any
    instance job ends in a failed state.

    .. seealso::
        For more information on how to use this sensor, take a look at the
        guide:
        :ref:`howto/sensor:SsmRunCommandCompletedSensor`

    :param command_id: The ID of the AWS SSM Run Command.
    :param fail_on_nonzero_exit: If True (default), the sensor will fail when the command
        returns a non-zero exit code. If False, the sensor will complete successfully
        for both Success and Failed command statuses, allowing downstream tasks to handle
        exit codes. AWS-level failures (Cancelled, TimedOut) will still raise exceptions.
        (default: True)
    :param deferrable: If True, the sensor will operate in deferrable mode.
        This mode requires aiobotocore module to be installed.
        (default: False, but can be overridden in config file by setting
        default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status
        of the job. (default: 120)
    :param max_retries: Number of times before returning the current state.
        (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used.
        If running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default
        boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore
        client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES: tuple[str, ...] = (
        "Pending",
        "Delayed",
        "InProgress",
        "Cancelling",
    )
    FAILURE_STATES: tuple[str, ...] = ("Cancelled", "TimedOut", "Failed")
    SUCCESS_STATES: tuple[str, ...] = ("Success",)
    FAILURE_MESSAGE = "SSM run command sensor failed."

    aws_hook_class = SsmHook
    template_fields: Sequence[str] = aws_template_fields(
        "command_id",
    )

    def __init__(
        self,
        *,
        command_id,
        fail_on_nonzero_exit: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poke_interval: int = 120,
        max_retries: int = 75,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.command_id = command_id
        self.fail_on_nonzero_exit = fail_on_nonzero_exit
        self.deferrable = deferrable
        self.poke_interval = poke_interval
        self.max_retries = max_retries

    def poke(self, context: Context):
        response = self.hook.conn.list_command_invocations(CommandId=self.command_id)
        command_invocations = response.get("CommandInvocations", [])

        if not command_invocations:
            self.log.info(
                "No command invocations found",
                "command_id=%s yet, waiting...",
                self.command_id,
            )
            return False

        for invocation in command_invocations:
            state = invocation["Status"]

            if state in self.FAILURE_STATES:
                # Check if we should tolerate this failure
                if self.fail_on_nonzero_exit:
                    raise RuntimeError(self.FAILURE_MESSAGE)  # Traditional behavior

                # Only fail on AWS-level issues, tolerate command failures
                if SsmHook.is_aws_level_failure(state):
                    raise RuntimeError(f"SSM command {self.command_id} {state}")

                # Command failed but we're tolerating it
                self.log.info(
                    "Command invocation has status %s. Continuing due to fail_on_nonzero_exit=False",
                    state,
                )

            if state in self.INTERMEDIATE_STATES:
                return False

        return True

    def execute(self, context: Context):
        if self.deferrable:
            self.defer(
                trigger=SsmRunCommandTrigger(
                    command_id=self.command_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                    fail_on_nonzero_exit=self.fail_on_nonzero_exit,
                ),
                method_name="execute_complete",
            )

        else:
            super().execute(context=context)

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> None:
        event = validate_execute_complete_event(event)

        if event["status"] != "success":
            raise RuntimeError(f"Error while running run command: {event}")

        self.log.info("SSM run command `%s` completed.", event["command_id"])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import logging
import subprocess
from datetime import datetime
from tempfile import NamedTemporaryFile
from textwrap import dedent

import boto3
from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.operators.sagemaker import (
    SageMakerAutoMLOperator,
    SageMakerCreateExperimentOperator,
    SageMakerDeleteModelOperator,
    SageMakerModelOperator,
    SageMakerProcessingOperator,
    SageMakerRegisterModelVersionOperator,
    SageMakerTrainingOperator,
    SageMakerTransformOperator,
    SageMakerTuningOperator,
)
from airflow.providers.amazon.aws.sensors.sagemaker import (
    SageMakerAutoMLSensor,
    SageMakerProcessingSensor,
    SageMakerTrainingSensor,
    SageMakerTransformSensor,
    SageMakerTuningSensor,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder, prune_logs

logger = logging.getLogger(__name__)

DAG_ID = "example_sagemaker"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

# The URI of a Docker image for handling KNN model training.
# To find the URI of a free Amazon-provided image that can be used, substitute your
# desired region in the following link and find the URI under "Registry Path".
# https://docs.aws.amazon.com/sagemaker/latest/dg/ecr-us-east-1.html#knn-us-east-1.title
# This URI should be in the format of {12-digits}.dkr.ecr.{region}.amazonaws.com/knn
KNN_IMAGES_BY_REGION = {
    "us-east-1": "382416733822.dkr.ecr.us-east-1.amazonaws.com/knn:1",
    "us-west-2": "174872318107.dkr.ecr.us-west-2.amazonaws.com/knn:1",
}

SAMPLE_SIZE = 600

# This script will be the entrypoint for the docker image which will handle preprocessing the raw data
PREPROCESS_SCRIPT_TEMPLATE = dedent("""
    import numpy as np
    import pandas as pd

    def main():
        # Load the dataset from {input_path}/input.csv, split it into train/test
        # subsets, and write them to {output_path}/ for the Processing Operator.

        data = pd.read_csv('{input_path}/input.csv')

        # Split into test and train data
        data_train, data_test = np.split(
            data.sample(frac=1, random_state=np.random.RandomState()), [int(0.7 * len(data))]
        )

        # Remove the "answers" from the test set
        data_test.drop(['class'], axis=1, inplace=True)

        # Write the splits to disk
        data_train.to_csv('{output_path}/train.csv', index=False, header=False)
        data_test.to_csv('{output_path}/test.csv', index=False, header=False)

        print('Preprocessing Done.')

    if __name__ == "__main__":
        main()
""")


def _install_aws_cli_if_needed():
    """
    Check if the AWS CLI tool is installed and install it if needed.

    The AmazonLinux image has flip-flopped a couple of times on whether this is included in the base image
    or not, so to future-proof this we are going to check if it's installed and install if necessary.
    """
    check = subprocess.Popen(
        "aws --version",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _, stderr = check.communicate()

    if check.returncode == 0:
        logger.info("AWS CLI tool is installed.")
        return

    if "aws: not found" in str(stderr):
        logger.info("AWS CLI tool not found; installing.")
        subprocess.Popen(
            """
                curl "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "awscliv2.zip"
                unzip awscliv2.zip
                sudo ./aws/install
            """,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()


def _create_ecr_repository(repo_name):
    execution_role_arn = boto3.client("sts").get_caller_identity()["Arn"]
    access_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "Allow access to the system test execution role",
                "Effect": "Allow",
                "Principal": {"AWS": execution_role_arn},
                "Action": "ecr:*",
            }
        ],
    }

    client = boto3.client("ecr")
    repo = client.create_repository(repositoryName=repo_name)["repository"]
    client.set_repository_policy(repositoryName=repo["repositoryName"], policyText=json.dumps(access_policy))

    return repo["repositoryUri"]


def _build_and_upload_docker_image(preprocess_script, repository_uri):
    """
    We need a Docker image with the following requirements:
      - Has numpy, pandas, requests, and boto3 installed
      - Has our data preprocessing script mounted and set as the entry point
    """
    with NamedTemporaryFile(mode="w+t") as preprocessing_script, NamedTemporaryFile(mode="w+t") as dockerfile:
        preprocessing_script.write(preprocess_script)
        preprocessing_script.flush()

        dockerfile.write(
            f"""
            FROM public.ecr.aws/amazonlinux/amazonlinux
            COPY {preprocessing_script.name.split("/")[2]} /preprocessing.py
            RUN yum install python3 pip -y
            RUN pip3 install boto3 pandas requests
            CMD [ "python3", "/preprocessing.py"]
            """
        )
        dockerfile.flush()

        ecr_region = repository_uri.split(".")[3]

        docker_build_and_push_commands = f"""
            # login to public ecr repo containing amazonlinux image (public login is always on us east 1)
            aws ecr-public get-login-password --region us-east-1 |
            docker login --username AWS --password-stdin public.ecr.aws &&
            docker build --platform=linux/amd64 -f {dockerfile.name} -t {repository_uri} /tmp &&

            # login again, this time to the private repo we created to hold that specific image
            aws ecr get-login-password --region {ecr_region} |
            docker login --username AWS --password-stdin {repository_uri} &&
            docker push {repository_uri}
            """
        logger.info("building and uploading docker image for preprocessing...")
        docker_build = subprocess.Popen(
            docker_build_and_push_commands,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _, stderr = docker_build.communicate()
        if docker_build.returncode != 0:
            raise RuntimeError(
                "Failed to prepare docker image for the preprocessing job.\n"
                "The following error happened while executing the sequence of bash commands:\n"
                f"{stderr.decode()}"
            )


def generate_data() -> str:
    """generates a very simple csv dataset with headers"""
    content = "class,x,y\n"  # headers
    for i in range(SAMPLE_SIZE):
        content += f"{i % 100},{i},{SAMPLE_SIZE - i}\n"
    return content


@task
def set_up(env_id, role_arn):
    bucket_name = f"{env_id}-sagemaker-example"
    ecr_repository_name = f"{env_id}-repo"
    model_name = f"{env_id}-KNN-model"
    processing_job_name = f"{env_id}-processing"
    training_job_name = f"{env_id}-train"
    transform_job_name = f"{env_id}-transform"
    tuning_job_name = f"{env_id}-tune"
    model_package_group_name = f"{env_id}-group"
    auto_ml_job_name = f"{env_id}-automl"
    experiment_name = f"{env_id}-experiment"

    input_data_S3_key = f"{env_id}/processed-input-data"
    prediction_output_s3_key = f"{env_id}/transform"
    processing_local_input_path = "/opt/ml/processing/input"
    processing_local_output_path = "/opt/ml/processing/output"
    raw_data_s3_key = f"{env_id}/preprocessing/input.csv"
    training_output_s3_key = f"{env_id}/results"

    ecr_repository_uri = _create_ecr_repository(ecr_repository_name)
    region = boto3.session.Session().region_name
    try:
        knn_image_uri = KNN_IMAGES_BY_REGION[region]
    except KeyError:
        raise KeyError(
            f"Region name {region} does not have a known KNN "
            f"Image URI.  Please add the region and URI following "
            f"the directions at the top of the system testfile "
        )

    resource_config = {
        "InstanceCount": 1,
        "InstanceType": "ml.m5.large",
        "VolumeSizeInGB": 1,
    }
    input_data_uri = f"s3://{bucket_name}/{raw_data_s3_key}"
    processing_config = {
        "ProcessingJobName": processing_job_name,
        "ProcessingInputs": [
            {
                "InputName": "input",
                "AppManaged": False,
                "S3Input": {
                    "S3Uri": input_data_uri,
                    "LocalPath": processing_local_input_path,
                    "S3DataType": "S3Prefix",
                    "S3InputMode": "File",
                    "S3DataDistributionType": "FullyReplicated",
                    "S3CompressionType": "None",
                },
            },
        ],
        "ProcessingOutputConfig": {
            "Outputs": [
                {
                    "OutputName": "output",
                    "S3Output": {
                        "S3Uri": f"s3://{bucket_name}/{input_data_S3_key}",
                        "LocalPath": processing_local_output_path,
                        "S3UploadMode": "EndOfJob",
                    },
                    "AppManaged": False,
                }
            ]
        },
        "ProcessingResources": {
            "ClusterConfig": resource_config,
        },
        "StoppingCondition": {"MaxRuntimeInSeconds": 600},
        "AppSpecification": {
            "ImageUri": ecr_repository_uri,
        },
        "RoleArn": role_arn,
    }

    training_data_source = {
        "CompressionType": "None",
        "ContentType": "text/csv",
        "DataSource": {
            "S3DataSource": {
                "S3DataDistributionType": "FullyReplicated",
                "S3DataType": "S3Prefix",
                "S3Uri": f"s3://{bucket_name}/{input_data_S3_key}/train.csv",
            }
        },
    }
    training_config = {
        "AlgorithmSpecification": {
            "TrainingImage": knn_image_uri,
            "TrainingInputMode": "File",
        },
        "HyperParameters": {
            "predictor_type": "classifier",
            "feature_dim": "2",
            "k": "3",
            "sample_size": str(SAMPLE_SIZE),
        },
        "InputDataConfig": [
            {
                "ChannelName": "train",
                **training_data_source,
            }
        ],
        "OutputDataConfig": {"S3OutputPath": f"s3://{bucket_name}/{training_output_s3_key}/"},
        "ExperimentConfig": {"ExperimentName": experiment_name},
        "ResourceConfig": resource_config,
        "RoleArn": role_arn,
        "StoppingCondition": {"MaxRuntimeInSeconds": 600},
        "TrainingJobName": training_job_name,
    }
    model_trained_weights = (
        f"s3://{bucket_name}/{training_output_s3_key}/{training_job_name}/output/model.tar.gz"
    )
    model_config = {
        "ExecutionRoleArn": role_arn,
        "ModelName": model_name,
        "PrimaryContainer": {
            "Mode": "SingleModel",
            "Image": knn_image_uri,
            "ModelDataUrl": model_trained_weights,
        },
    }
    tuning_config = {
        "HyperParameterTuningJobName": tuning_job_name,
        "HyperParameterTuningJobConfig": {
            "Strategy": "Bayesian",
            "HyperParameterTuningJobObjective": {
                "MetricName": "test:accuracy",
                "Type": "Maximize",
            },
            "ResourceLimits": {
                "MaxNumberOfTrainingJobs": 10,
                "MaxParallelTrainingJobs": 10,
            },
            "ParameterRanges": {
                "CategoricalParameterRanges": [],
                "IntegerParameterRanges": [
                    # Set the min and max values of the hyperparameters you want to tune.
                    {
                        "Name": "k",
                        "MinValue": "1",
                        "MaxValue": str(SAMPLE_SIZE),
                    },
                    {
                        "Name": "sample_size",
                        "MinValue": "1",
                        "MaxValue": str(SAMPLE_SIZE),
                    },
                ],
            },
        },
        "TrainingJobDefinition": {
            "StaticHyperParameters": {
                "predictor_type": "classifier",
                "feature_dim": "2",
            },
            "AlgorithmSpecification": {"TrainingImage": knn_image_uri, "TrainingInputMode": "File"},
            "InputDataConfig": [
                {
                    "ChannelName": "train",
                    **training_data_source,
                },
                {
                    "ChannelName": "test",
                    **training_data_source,
                },
            ],
            "OutputDataConfig": {"S3OutputPath": f"s3://{bucket_name}/{training_output_s3_key}"},
            "ResourceConfig": resource_config,
            "RoleArn": role_arn,
            "StoppingCondition": {"MaxRuntimeInSeconds": 600},
        },
    }
    transform_config = {
        "TransformJobName": transform_job_name,
        "TransformInput": {
            "DataSource": {
                "S3DataSource": {
                    "S3DataType": "S3Prefix",
                    "S3Uri": f"s3://{bucket_name}/{input_data_S3_key}/test.csv",
                }
            },
            "SplitType": "Line",
            "ContentType": "text/csv",
        },
        "TransformOutput": {"S3OutputPath": f"s3://{bucket_name}/{prediction_output_s3_key}"},
        "TransformResources": {
            "InstanceCount": 1,
            "InstanceType": "ml.m5.large",
        },
        "ModelName": model_name,
    }

    preprocess_script = PREPROCESS_SCRIPT_TEMPLATE.format(
        input_path=processing_local_input_path, output_path=processing_local_output_path
    )
    _install_aws_cli_if_needed()
    _build_and_upload_docker_image(preprocess_script, ecr_repository_uri)

    from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

    if AIRFLOW_V_3_0_PLUS:
        from airflow.sdk import get_current_context
    else:
        from airflow.operators.python import get_current_context  # type: ignore[attr-defined,no-redef]

    ti = get_current_context()["ti"]
    ti.xcom_push(key="docker_image", value=ecr_repository_uri)
    ti.xcom_push(key="bucket_name", value=bucket_name)
    ti.xcom_push(key="raw_data_s3_key", value=raw_data_s3_key)
    ti.xcom_push(key="ecr_repository_name", value=ecr_repository_name)
    ti.xcom_push(key="processing_config", value=processing_config)
    ti.xcom_push(key="processing_job_name", value=processing_job_name)
    ti.xcom_push(key="input_data_uri", value=input_data_uri)
    ti.xcom_push(key="output_data_uri", value=f"s3://{bucket_name}/{training_output_s3_key}")
    ti.xcom_push(key="training_config", value=training_config)
    ti.xcom_push(key="training_job_name", value=training_job_name)
    ti.xcom_push(key="model_package_group_name", value=model_package_group_name)
    ti.xcom_push(key="auto_ml_job_name", value=auto_ml_job_name)
    ti.xcom_push(key="experiment_name", value=experiment_name)
    ti.xcom_push(key="model_config", value=model_config)
    ti.xcom_push(key="model_name", value=model_name)
    ti.xcom_push(key="inference_code_image", value=knn_image_uri)
    ti.xcom_push(key="model_trained_weights", value=model_trained_weights)
    ti.xcom_push(key="tuning_config", value=tuning_config)
    ti.xcom_push(key="tuning_job_name", value=tuning_job_name)
    ti.xcom_push(key="transform_config", value=transform_config)
    ti.xcom_push(key="transform_job_name", value=transform_job_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def stop_automl_job(job_name: str):
    try:
        logger.info("Stopping AutoML job: %s", job_name)
        boto3.client("sagemaker").stop_auto_ml_job(AutoMLJobName=job_name)
    except ClientError as e:
        # If the job has already completed, boto will raise a ValidationException.
        # In this case, consider that a successful result.
        if (
            e.response["Error"]["Code"] == "ValidationException"
            and "already reached a terminal state" in e.response["Error"]["Message"]
        ):
            logger.info("AutoML job %s already completed.", job_name)
        else:
            raise e


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_ecr_repository(repository_name):
    client = boto3.client("ecr")

    # All images must be removed from the repo before it can be deleted.
    image_ids = client.list_images(repositoryName=repository_name)["imageIds"]
    client.batch_delete_image(
        repositoryName=repository_name,
        imageIds=[{"imageDigest": image["imageDigest"]} for image in image_ids],
    )
    client.delete_repository(repositoryName=repository_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_model_group(group_name, model_version_arn):
    sgmk_client = boto3.client("sagemaker")
    # need to destroy model registered in group first
    sgmk_client.delete_model_package(ModelPackageName=model_version_arn)
    sgmk_client.delete_model_package_group(ModelPackageGroupName=group_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_experiments(experiment_names):
    sgmk_client = boto3.client("sagemaker")
    for experiment in experiment_names:
        trials = sgmk_client.list_trials(ExperimentName=experiment)
        trials_names = [s["TrialName"] for s in trials["TrialSummaries"]]
        for trial in trials_names:
            components = sgmk_client.list_trial_components(TrialName=trial)
            components_names = [s["TrialComponentName"] for s in components["TrialComponentSummaries"]]
            for component in components_names:
                sgmk_client.disassociate_trial_component(TrialComponentName=component, TrialName=trial)
                sgmk_client.delete_trial_component(TrialComponentName=component)
            sgmk_client.delete_trial(TrialName=trial)
        sgmk_client.delete_experiment(ExperimentName=experiment)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_docker_image(image_name):
    docker_build = subprocess.Popen(
        f"docker rmi {image_name}",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _, stderr = docker_build.communicate()
    if docker_build.returncode != 0:
        logger.error(
            "Failed to delete local docker image. "
            "Run 'docker images' to see if you need to clean it yourself.\nerror message: %s",
            stderr,
        )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]

    test_setup = set_up(
        env_id=env_id,
        role_arn=test_context[ROLE_ARN_KEY],
    )

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=test_setup["bucket_name"],
    )

    upload_dataset = S3CreateObjectOperator(
        task_id="upload_dataset",
        s3_bucket=test_setup["bucket_name"],
        s3_key=test_setup["raw_data_s3_key"],
        data=generate_data(),
        replace=True,
    )

    # [START howto_operator_sagemaker_auto_ml]
    automl = SageMakerAutoMLOperator(
        task_id="auto_ML",
        job_name=test_setup["auto_ml_job_name"],
        s3_input=test_setup["input_data_uri"],
        target_attribute="class",
        s3_output=test_setup["output_data_uri"],
        role_arn=test_context[ROLE_ARN_KEY],
        time_limit=30,  # will stop the job before it can do anything, but it's not the point here
    )
    # [END howto_operator_sagemaker_auto_ml]
    automl.wait_for_completion = False  # just to be able to test the sensor next

    # [START howto_sensor_sagemaker_auto_ml]
    await_automl = SageMakerAutoMLSensor(job_name=test_setup["auto_ml_job_name"], task_id="await_auto_ML")
    # [END howto_sensor_sagemaker_auto_ml]
    await_automl.poke_interval = 10

    # [START howto_operator_sagemaker_experiment]
    create_experiment = SageMakerCreateExperimentOperator(
        task_id="create_experiment", name=test_setup["experiment_name"]
    )
    # [END howto_operator_sagemaker_experiment]

    # [START howto_operator_sagemaker_processing]
    preprocess_raw_data = SageMakerProcessingOperator(
        task_id="preprocess_raw_data",
        config=test_setup["processing_config"],
    )

    # SageMakerProcessingOperator waits by default, setting as False to test the Sensor below.
    preprocess_raw_data.wait_for_completion = False

    # [END howto_operator_sagemaker_processing]

    # [START howto_sensor_sagemaker_processing]
    await_preprocess = SageMakerProcessingSensor(
        task_id="await_preprocess", job_name=test_setup["processing_job_name"]
    )
    # [END howto_sensor_sagemaker_processing]

    # [START howto_operator_sagemaker_training]
    train_model = SageMakerTrainingOperator(
        task_id="train_model",
        config=test_setup["training_config"],
    )
    # [END howto_operator_sagemaker_training]

    # SageMakerTrainingOperator waits by default, setting as False to test the Sensor below.
    train_model.wait_for_completion = False

    # [START howto_sensor_sagemaker_training]
    await_training = SageMakerTrainingSensor(
        task_id="await_training",
        job_name=test_setup["training_job_name"],
    )
    # [END howto_sensor_sagemaker_training]

    # [START howto_operator_sagemaker_model]
    create_model = SageMakerModelOperator(
        task_id="create_model",
        config=test_setup["model_config"],
    )
    # [END howto_operator_sagemaker_model]

    # [START howto_operator_sagemaker_register]
    register_model = SageMakerRegisterModelVersionOperator(
        task_id="register_model",
        image_uri=test_setup["inference_code_image"],
        model_url=test_setup["model_trained_weights"],
        package_group_name=test_setup["model_package_group_name"],
    )
    # [END howto_operator_sagemaker_register]

    # [START howto_operator_sagemaker_tuning]
    tune_model = SageMakerTuningOperator(
        task_id="tune_model",
        config=test_setup["tuning_config"],
    )
    # [END howto_operator_sagemaker_tuning]

    # SageMakerTuningOperator waits by default, setting as False to test the Sensor below.
    tune_model.wait_for_completion = False

    # [START howto_sensor_sagemaker_tuning]
    await_tuning = SageMakerTuningSensor(
        task_id="await_tuning",
        job_name=test_setup["tuning_job_name"],
    )
    # [END howto_sensor_sagemaker_tuning]

    # [START howto_operator_sagemaker_transform]
    test_model = SageMakerTransformOperator(
        task_id="test_model",
        config=test_setup["transform_config"],
    )
    # [END howto_operator_sagemaker_transform]

    # SageMakerTransformOperator waits by default, setting as False to test the Sensor below.
    test_model.wait_for_completion = False

    # [START howto_sensor_sagemaker_transform]
    await_transform = SageMakerTransformSensor(
        task_id="await_transform",
        job_name=test_setup["transform_job_name"],
    )
    # [END howto_sensor_sagemaker_transform]

    # [START howto_operator_sagemaker_delete_model]
    delete_model = SageMakerDeleteModelOperator(
        task_id="delete_model",
        config={"ModelName": test_setup["model_name"]},
    )
    # [END howto_operator_sagemaker_delete_model]
    delete_model.trigger_rule = TriggerRule.ALL_DONE

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=test_setup["bucket_name"],
        force_delete=True,
    )

    log_cleanup = prune_logs(
        [
            # Format: ('log group name', 'log stream prefix')
            ("/aws/sagemaker/ProcessingJobs", env_id),
            ("/aws/sagemaker/TrainingJobs", env_id),
            ("/aws/sagemaker/TransformJobs", env_id),
        ]
    )

    chain(
        # TEST SETUP
        test_context,
        test_setup,
        create_bucket,
        upload_dataset,
        # TEST BODY
        automl,
        await_automl,
        create_experiment,
        preprocess_raw_data,
        await_preprocess,
        train_model,
        await_training,
        create_model,
        register_model,
        tune_model,
        await_tuning,
        test_model,
        await_transform,
        # TEST TEARDOWN
        delete_ecr_repository(test_setup["ecr_repository_name"]),
        delete_model_group(test_setup["model_package_group_name"], register_model.output),
        delete_model,
        delete_bucket,
        stop_automl_job(test_setup["auto_ml_job_name"]),
        delete_experiments(
            [
                test_setup["experiment_name"],
                f"{test_setup['auto_ml_job_name']}-aws-auto-ml-job",
                f"{test_setup['tuning_job_name']}-aws-tuning-job",
            ]
        ),
        delete_docker_image(test_setup["docker_image"]),
        log_cleanup,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import time
from datetime import datetime, timezone
from unittest import mock
from unittest.mock import patch

import pytest
from botocore.exceptions import ClientError
from dateutil.tz import tzlocal
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.logs import AwsLogsHook
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.hooks.sagemaker import (
    LogState,
    SageMakerHook,
    secondary_training_status_changed,
    secondary_training_status_message,
)
from airflow.providers.common.compat.sdk import AirflowException

role = "arn:aws:iam:role/test-role"

path = "local/data"
bucket = "test-bucket"
key = "test/data"
data_url = f"s3://{bucket}/{key}"

job_name = "test-job"
model_name = "test-model"
config_name = "test-endpoint-config"
endpoint_name = "test-endpoint"

image = "test-image"
test_arn_return = {"Arn": "testarn"}
output_url = f"s3://{bucket}/test/output"

create_training_params = {
    "AlgorithmSpecification": {"TrainingImage": image, "TrainingInputMode": "File"},
    "RoleArn": role,
    "OutputDataConfig": {"S3OutputPath": output_url},
    "ResourceConfig": {"InstanceCount": 2, "InstanceType": "ml.c6g.8xlarge", "VolumeSizeInGB": 50},
    "TrainingJobName": job_name,
    "HyperParameters": {"k": "10", "feature_dim": "784", "mini_batch_size": "500", "force_dense": "True"},
    "StoppingCondition": {"MaxRuntimeInSeconds": 60 * 60},
    "InputDataConfig": [
        {
            "ChannelName": "train",
            "DataSource": {
                "S3DataSource": {
                    "S3DataType": "S3Prefix",
                    "S3Uri": data_url,
                    "S3DataDistributionType": "FullyReplicated",
                }
            },
            "CompressionType": "None",
            "RecordWrapperType": "None",
        },
        {
            "ChannelName": "train_fs",
            "DataSource": {
                "FileSystemDataSource": {
                    "DirectoryPath": "/tmp",
                    "FileSystemAccessMode": "ro",
                    "FileSystemId": "fs-abc",
                    "FileSystemType": "FSxLustre",
                }
            },
            "CompressionType": "None",
            "RecordWrapperType": "None",
        },
    ],
}

create_tuning_params = {
    "HyperParameterTuningJobName": job_name,
    "HyperParameterTuningJobConfig": {
        "Strategy": "Bayesian",
        "HyperParameterTuningJobObjective": {"Type": "Maximize", "MetricName": "test_metric"},
        "ResourceLimits": {"MaxNumberOfTrainingJobs": 123, "MaxParallelTrainingJobs": 123},
        "ParameterRanges": {
            "IntegerParameterRanges": [
                {"Name": "k", "MinValue": "2", "MaxValue": "10"},
            ]
        },
    },
    "TrainingJobDefinition": {
        "StaticHyperParameters": create_training_params["HyperParameters"],
        "AlgorithmSpecification": create_training_params["AlgorithmSpecification"],
        "RoleArn": "string",
        "InputDataConfig": create_training_params["InputDataConfig"],
        "OutputDataConfig": create_training_params["OutputDataConfig"],
        "ResourceConfig": create_training_params["ResourceConfig"],
        "StoppingCondition": dict(MaxRuntimeInSeconds=60 * 60),
    },
}

create_transform_params = {
    "TransformJobName": job_name,
    "ModelName": model_name,
    "BatchStrategy": "MultiRecord",
    "TransformInput": {"DataSource": {"S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": data_url}}},
    "TransformOutput": {
        "S3OutputPath": output_url,
    },
    "TransformResources": {"InstanceType": "ml.m6g.xlarge", "InstanceCount": 123},
}

create_transform_params_fs = {
    "TransformJobName": job_name,
    "ModelName": model_name,
    "BatchStrategy": "MultiRecord",
    "TransformInput": {
        "DataSource": {
            "FileSystemDataSource": {
                "DirectoryPath": "/tmp",
                "FileSystemAccessMode": "ro",
                "FileSystemId": "fs-abc",
                "FileSystemType": "FSxLustre",
            }
        }
    },
    "TransformOutput": {
        "S3OutputPath": output_url,
    },
    "TransformResources": {"InstanceType": "ml.m6g.xlarge", "InstanceCount": 123},
}

create_model_params = {
    "ModelName": model_name,
    "PrimaryContainer": {
        "Image": image,
        "ModelDataUrl": output_url,
    },
    "ExecutionRoleArn": role,
}

create_endpoint_config_params = {
    "EndpointConfigName": config_name,
    "ProductionVariants": [
        {
            "VariantName": "AllTraffic",
            "ModelName": model_name,
            "InitialInstanceCount": 1,
            "InstanceType": "ml.c6g.xlarge",
        }
    ],
}

create_endpoint_params = {"EndpointName": endpoint_name, "EndpointConfigName": config_name}

update_endpoint_params = create_endpoint_params

DESCRIBE_TRAINING_COMPLETED_RETURN = {
    "TrainingJobStatus": "Completed",
    "ResourceConfig": {"InstanceCount": 1, "InstanceType": "ml.c6g.xlarge", "VolumeSizeInGB": 10},
    "TrainingStartTime": datetime(2018, 2, 17, 7, 15, 0, 103000),
    "TrainingEndTime": datetime(2018, 2, 17, 7, 19, 34, 953000),
    "ResponseMetadata": {
        "HTTPStatusCode": 200,
    },
}

DESCRIBE_TRAINING_INPROGRESS_RETURN = dict(DESCRIBE_TRAINING_COMPLETED_RETURN)
DESCRIBE_TRAINING_INPROGRESS_RETURN.update({"TrainingJobStatus": "InProgress"})

DESCRIBE_TRAINING_FAILED_RETURN = dict(DESCRIBE_TRAINING_COMPLETED_RETURN)
DESCRIBE_TRAINING_FAILED_RETURN.update({"TrainingJobStatus": "Failed", "FailureReason": "Unknown"})

DESCRIBE_TRAINING_STOPPING_RETURN = dict(DESCRIBE_TRAINING_COMPLETED_RETURN)
DESCRIBE_TRAINING_STOPPING_RETURN.update({"TrainingJobStatus": "Stopping"})

message = "message"
status = "status"
SECONDARY_STATUS_DESCRIPTION_1 = {
    "SecondaryStatusTransitions": [{"StatusMessage": message, "Status": status}]
}
SECONDARY_STATUS_DESCRIPTION_2 = {
    "SecondaryStatusTransitions": [{"StatusMessage": "different message", "Status": status}]
}

DEFAULT_LOG_STREAMS = {"logStreams": [{"logStreamName": job_name + "/xxxxxxxxx"}]}
LIFECYCLE_LOG_STREAMS = [
    DEFAULT_LOG_STREAMS,
    DEFAULT_LOG_STREAMS,
    DEFAULT_LOG_STREAMS,
    DEFAULT_LOG_STREAMS,
    DEFAULT_LOG_STREAMS,
    DEFAULT_LOG_STREAMS,
]

DEFAULT_LOG_EVENTS = [
    {"nextForwardToken": None, "events": [{"timestamp": 1, "message": "hi there #1"}]},
    {"nextForwardToken": None, "events": []},
]
STREAM_LOG_EVENTS = [
    {"nextForwardToken": None, "events": [{"timestamp": 1, "message": "hi there #1"}]},
    {"nextForwardToken": None, "events": []},
    {
        "nextForwardToken": None,
        "events": [{"timestamp": 1, "message": "hi there #1"}, {"timestamp": 2, "message": "hi there #2"}],
    },
    {"nextForwardToken": None, "events": []},
    {
        "nextForwardToken": None,
        "events": [
            {"timestamp": 2, "message": "hi there #2"},
            {"timestamp": 2, "message": "hi there #2a"},
            {"timestamp": 3, "message": "hi there #3"},
        ],
    },
    {"nextForwardToken": None, "events": []},
]

test_evaluation_config = {
    "Image": image,
    "Role": role,
    "S3Operations": {
        "S3CreateBucket": [{"Bucket": bucket}],
        "S3Upload": [{"Path": path, "Bucket": bucket, "Key": key, "Tar": False}],
    },
}


class TestSageMakerHook:
    @mock.patch.object(AwsLogsHook, "get_log_events")
    def test_multi_stream_iter(self, mock_log_stream):
        event = {"timestamp": 1}
        mock_log_stream.side_effect = [iter([event]), iter([]), None]
        hook = SageMakerHook()
        event_iter = hook.multi_stream_iter("log", [None, None, None])
        assert next(event_iter) == (0, event)

    @mock.patch.object(S3Hook, "create_bucket")
    @mock.patch.object(S3Hook, "load_file")
    def test_configure_s3_resources(self, mock_load_file, mock_create_bucket):
        hook = SageMakerHook()
        evaluation_result = {"Image": image, "Role": role}
        hook.configure_s3_resources(test_evaluation_config)
        assert test_evaluation_config == evaluation_result
        mock_create_bucket.assert_called_once_with(bucket_name=bucket)
        mock_load_file.assert_called_once_with(path, key, bucket)

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(S3Hook, "check_for_key")
    @mock.patch.object(S3Hook, "check_for_bucket")
    @mock.patch.object(S3Hook, "check_for_prefix")
    def test_check_s3_url(self, mock_check_prefix, mock_check_bucket, mock_check_key, mock_client):
        mock_client.return_value = None
        hook = SageMakerHook()
        mock_check_bucket.side_effect = [False, True, True, True]
        mock_check_key.side_effect = [False, True, False]
        mock_check_prefix.side_effect = [False, True, True]
        with pytest.raises(AirflowException):
            hook.check_s3_url(data_url)
        with pytest.raises(AirflowException):
            hook.check_s3_url(data_url)
        assert hook.check_s3_url(data_url) is True
        assert hook.check_s3_url(data_url) is True

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(SageMakerHook, "check_s3_url")
    def test_check_valid_training(self, mock_check_url, mock_client):
        mock_client.return_value = None
        hook = SageMakerHook()
        hook.check_training_config(create_training_params)
        mock_check_url.assert_called_once_with(data_url)

        # InputDataConfig is optional, verify if check succeeds without InputDataConfig
        create_training_params_no_inputdataconfig = create_training_params.copy()
        create_training_params_no_inputdataconfig.pop("InputDataConfig")
        hook.check_training_config(create_training_params_no_inputdataconfig)

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(SageMakerHook, "check_s3_url")
    def test_check_valid_tuning(self, mock_check_url, mock_client):
        mock_client.return_value = None
        hook = SageMakerHook()
        hook.check_tuning_config(create_tuning_params)
        mock_check_url.assert_called_once_with(data_url)

    def test_conn(self):
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        assert hook.aws_conn_id == "sagemaker_test_conn_id"

    @mock.patch.object(SageMakerHook, "check_training_config")
    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_training_job(self, mock_client, mock_check_training):
        mock_check_training.return_value = True
        mock_session = mock.Mock()
        attrs = {"create_training_job.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_training_job(
            create_training_params, wait_for_completion=False, print_log=False
        )
        mock_session.create_training_job.assert_called_once_with(**create_training_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "check_training_config")
    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch("time.sleep", return_value=None)
    def test_training_ends_with_wait(self, _, mock_client, mock_check_training):
        mock_check_training.return_value = True
        mock_session = mock.Mock()
        attrs = {
            "create_training_job.return_value": test_arn_return,
            "describe_training_job.side_effect": [
                DESCRIBE_TRAINING_INPROGRESS_RETURN,
                DESCRIBE_TRAINING_STOPPING_RETURN,
                DESCRIBE_TRAINING_COMPLETED_RETURN,
            ],
        }
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id_1")
        hook.create_training_job(
            create_training_params, wait_for_completion=True, print_log=False, check_interval=0
        )
        assert mock_session.describe_training_job.call_count == 3

    @mock.patch.object(SageMakerHook, "check_training_config")
    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch("time.sleep", return_value=None)
    def test_training_throws_error_when_failed_with_wait(self, _, mock_client, mock_check_training):
        mock_check_training.return_value = True
        mock_session = mock.Mock()
        attrs = {
            "create_training_job.return_value": test_arn_return,
            "describe_training_job.side_effect": [
                DESCRIBE_TRAINING_INPROGRESS_RETURN,
                DESCRIBE_TRAINING_STOPPING_RETURN,
                DESCRIBE_TRAINING_FAILED_RETURN,
                DESCRIBE_TRAINING_COMPLETED_RETURN,
            ],
        }
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id_1")
        with pytest.raises(AirflowException):
            hook.create_training_job(
                create_training_params,
                wait_for_completion=True,
                print_log=False,
                check_interval=0,
            )
        assert mock_session.describe_training_job.call_count == 3

    @mock.patch.object(SageMakerHook, "check_tuning_config")
    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_tuning_job(self, mock_client, mock_check_tuning_config):
        mock_session = mock.Mock()
        attrs = {"create_hyper_parameter_tuning_job.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_tuning_job(create_tuning_params, wait_for_completion=False)
        mock_session.create_hyper_parameter_tuning_job.assert_called_once_with(**create_tuning_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "check_s3_url")
    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_transform_job(self, mock_client, mock_check_url):
        mock_check_url.return_value = True
        mock_session = mock.Mock()
        attrs = {"create_transform_job.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_transform_job(create_transform_params, wait_for_completion=False)
        mock_session.create_transform_job.assert_called_once_with(**create_transform_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_transform_job_fs(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"create_transform_job.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_transform_job(create_transform_params_fs, wait_for_completion=False)
        mock_session.create_transform_job.assert_called_once_with(**create_transform_params_fs)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_model(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"create_model.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_model(create_model_params)
        mock_session.create_model.assert_called_once_with(**create_model_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_endpoint_config(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"create_endpoint_config.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_endpoint_config(create_endpoint_config_params)
        mock_session.create_endpoint_config.assert_called_once_with(**create_endpoint_config_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_create_endpoint(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"create_endpoint.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.create_endpoint(create_endpoint_params, wait_for_completion=False)
        mock_session.create_endpoint.assert_called_once_with(**create_endpoint_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_update_endpoint(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"update_endpoint.return_value": test_arn_return}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.update_endpoint(update_endpoint_params, wait_for_completion=False)
        mock_session.update_endpoint.assert_called_once_with(**update_endpoint_params)
        assert response == test_arn_return

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_training_job(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"describe_training_job.return_value": "InProgress"}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_training_job(job_name)
        mock_session.describe_training_job.assert_called_once_with(TrainingJobName=job_name)
        assert response == "InProgress"

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_tuning_job(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"describe_hyper_parameter_tuning_job.return_value": "InProgress"}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_tuning_job(job_name)
        mock_session.describe_hyper_parameter_tuning_job.assert_called_once_with(
            HyperParameterTuningJobName=job_name
        )
        assert response == "InProgress"

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_transform_job(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"describe_transform_job.return_value": "InProgress"}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_transform_job(job_name)
        mock_session.describe_transform_job.assert_called_once_with(TransformJobName=job_name)
        assert response == "InProgress"

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_model(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"describe_model.return_value": model_name}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_model(model_name)
        mock_session.describe_model.assert_called_once_with(ModelName=model_name)
        assert response == model_name

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_endpoint_config(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"describe_endpoint_config.return_value": config_name}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_endpoint_config(config_name)
        mock_session.describe_endpoint_config.assert_called_once_with(EndpointConfigName=config_name)
        assert response == config_name

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_endpoint(self, mock_client):
        mock_session = mock.Mock()
        attrs = {"describe_endpoint.return_value": "InProgress"}
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_endpoint(endpoint_name)
        mock_session.describe_endpoint.assert_called_once_with(EndpointName=endpoint_name)
        assert response == "InProgress"

    def test_secondary_training_status_changed_true(self):
        changed = secondary_training_status_changed(
            SECONDARY_STATUS_DESCRIPTION_1, SECONDARY_STATUS_DESCRIPTION_2
        )
        assert changed

    def test_secondary_training_status_changed_false(self):
        changed = secondary_training_status_changed(
            SECONDARY_STATUS_DESCRIPTION_1, SECONDARY_STATUS_DESCRIPTION_1
        )
        assert not changed

    def test_secondary_training_status_message_status_changed(self):
        now = datetime.now(tzlocal())
        SECONDARY_STATUS_DESCRIPTION_1["LastModifiedTime"] = now
        expected_time = now.astimezone(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        expected = f"{expected_time} {status} - {message}"
        assert (
            secondary_training_status_message(SECONDARY_STATUS_DESCRIPTION_1, SECONDARY_STATUS_DESCRIPTION_2)
            == expected
        )

    @mock.patch.object(AwsLogsHook, "conn")
    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(time, "monotonic")
    def test_describe_training_job_with_logs_in_progress(self, mock_time, mock_client, mock_log_client):
        mock_session = mock.Mock()
        mock_log_session = mock.Mock()
        attrs = {"describe_training_job.return_value": DESCRIBE_TRAINING_COMPLETED_RETURN}
        log_attrs = {
            "describe_log_streams.side_effect": LIFECYCLE_LOG_STREAMS,
            "get_log_events.side_effect": STREAM_LOG_EVENTS,
        }
        mock_time.return_value = 50
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        mock_log_session.configure_mock(**log_attrs)
        mock_log_client.return_value = mock_log_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_training_job_with_log(
            job_name=job_name,
            positions={},
            stream_names=[],
            instance_count=1,
            state=LogState.WAIT_IN_PROGRESS,
            last_description={},
            last_describe_job_call=0,
        )
        assert response == (LogState.JOB_COMPLETE, {}, 50)

    @pytest.mark.parametrize("log_state", [LogState.JOB_COMPLETE, LogState.COMPLETE])
    @mock.patch.object(AwsLogsHook, "conn")
    @mock.patch.object(SageMakerHook, "get_conn")
    def test_describe_training_job_with_complete_states(self, mock_client, mock_log_client, log_state):
        mock_session = mock.Mock()
        mock_log_session = mock.Mock()
        attrs = {"describe_training_job.return_value": DESCRIBE_TRAINING_COMPLETED_RETURN}
        log_attrs = {
            "describe_log_streams.side_effect": LIFECYCLE_LOG_STREAMS,
            "get_log_events.side_effect": STREAM_LOG_EVENTS,
        }
        mock_session.configure_mock(**attrs)
        mock_client.return_value = mock_session
        mock_log_session.configure_mock(**log_attrs)
        mock_log_client.return_value = mock_log_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        response = hook.describe_training_job_with_log(
            job_name=job_name,
            positions={},
            stream_names=[],
            instance_count=1,
            state=log_state,
            last_description={},
            last_describe_job_call=0,
        )
        assert response == (LogState.COMPLETE, {}, 0)

    @mock.patch.object(SageMakerHook, "check_training_config")
    @mock.patch.object(AwsLogsHook, "conn")
    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(SageMakerHook, "describe_training_job_with_log")
    @mock.patch("time.sleep", return_value=None)
    def test_training_with_logs(self, _, mock_describe, mock_client, mock_log_client, mock_check_training):
        mock_check_training.return_value = True
        mock_describe.side_effect = [
            (LogState.WAIT_IN_PROGRESS, DESCRIBE_TRAINING_INPROGRESS_RETURN, 0),
            (LogState.JOB_COMPLETE, DESCRIBE_TRAINING_STOPPING_RETURN, 0),
            (LogState.COMPLETE, DESCRIBE_TRAINING_COMPLETED_RETURN, 0),
        ]
        mock_session = mock.Mock()
        mock_log_session = mock.Mock()
        attrs = {
            "create_training_job.return_value": test_arn_return,
            "describe_training_job.return_value": DESCRIBE_TRAINING_COMPLETED_RETURN,
        }
        log_attrs = {
            "describe_log_streams.side_effect": LIFECYCLE_LOG_STREAMS,
            "get_log_events.side_effect": STREAM_LOG_EVENTS,
        }
        mock_session.configure_mock(**attrs)
        mock_log_session.configure_mock(**log_attrs)
        mock_client.return_value = mock_session
        mock_log_client.return_value = mock_log_session
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id_1")
        hook.create_training_job(
            create_training_params, wait_for_completion=True, print_log=True, check_interval=0
        )
        assert mock_describe.call_count == 3
        assert mock_session.describe_training_job.call_count == 1

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_count_processing_jobs_by_name(self, mock_conn):
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        existing_job_name = "existing_job"
        mock_conn().list_processing_jobs.return_value = {
            "ProcessingJobSummaries": [{"ProcessingJobName": existing_job_name}]
        }
        ret = hook.count_processing_jobs_by_name(existing_job_name)
        assert ret == 1

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_count_processing_jobs_by_name_only_counts_actual_hits(self, mock_conn):
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        existing_job_name = "existing_job"
        mock_conn().list_processing_jobs.return_value = {
            "ProcessingJobSummaries": [
                {"ProcessingJobName": existing_job_name},
                {"ProcessingJobName": f"contains_but_does_not_start_with_{existing_job_name}"},
                {"ProcessingJobName": f"{existing_job_name}_with_different_suffix-123"},
            ]
        }
        ret = hook.count_processing_jobs_by_name(existing_job_name)
        assert ret == 1

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch("time.sleep", return_value=None)
    def test_count_processing_jobs_by_name_retries_on_throttle_exception(self, _, mock_conn):
        throttle_exception = ClientError(
            error_response={"Error": {"Code": "ThrottlingException"}}, operation_name="empty"
        )
        successful_result = {"ProcessingJobSummaries": [{"ProcessingJobName": "existing_job"}]}
        # Return a ThrottleException on the first call, then a mocked successful value the second.
        mock_conn().list_processing_jobs.side_effect = [throttle_exception, successful_result]
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")

        ret = hook.count_processing_jobs_by_name("existing_job")

        assert mock_conn().list_processing_jobs.call_count == 2
        assert ret == 1

    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch("time.sleep", return_value=None)
    def test_count_processing_jobs_by_name_fails_after_max_retries(self, _, mock_conn):
        mock_conn().list_processing_jobs.side_effect = ClientError(
            error_response={"Error": {"Code": "ThrottlingException"}}, operation_name="empty"
        )
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")
        retries = 3

        with pytest.raises(ClientError) as raised_exception:
            hook.count_processing_jobs_by_name("existing_job", retries=retries)

        assert mock_conn().list_processing_jobs.call_count == retries + 1
        assert raised_exception.value.response["Error"]["Code"] == "ThrottlingException"

    @mock.patch.object(SageMakerHook, "get_conn")
    def test_count_processing_jobs_by_name_job_not_exists_should_return_falsy(self, mock_conn):
        error_resp = {"Error": {"Code": "ResourceNotFound"}}
        mock_conn().list_processing_jobs.side_effect = ClientError(
            error_response=error_resp, operation_name="empty"
        )
        hook = SageMakerHook(aws_conn_id="sagemaker_test_conn_id")

        ret = hook.count_processing_jobs_by_name("existing_job")
        assert ret == 0

    @mock_aws
    def test_delete_model(self):
        hook = SageMakerHook()
        with patch.object(hook.conn, "delete_model") as mock_delete:
            hook.delete_model(model_name="test")
        mock_delete.assert_called_once_with(ModelName="test")

    @mock_aws
    def test_delete_model_when_not_exist(self):
        hook = SageMakerHook()
        with pytest.raises(ClientError) as raised_exception:
            hook.delete_model(model_name="test")
        ex = raised_exception.value
        assert ex.operation_name == "DeleteModel"
        assert ex.response["ResponseMetadata"]["HTTPStatusCode"] == 404

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_start_pipeline_returns_arn(self, mock_conn):
        mock_conn().start_pipeline_execution.return_value = {"PipelineExecutionArn": "hellotest"}

        hook = SageMakerHook(aws_conn_id="aws_default")
        params_dict = {"one": "1", "two": "2"}
        arn = hook.start_pipeline(pipeline_name="test_name", pipeline_params=params_dict)

        assert arn == "hellotest"

        args_passed = mock_conn().start_pipeline_execution.call_args.kwargs
        assert args_passed["PipelineName"] == "test_name"

        # check conversion to the weird format for passing parameters (list of tuples)
        assert len(args_passed["PipelineParameters"]) == 2
        for transformed_param in args_passed["PipelineParameters"]:
            assert "Name" in transformed_param.keys()
            assert "Value" in transformed_param.keys()
            # Name contains the key
            assert transformed_param["Name"] in params_dict.keys()
            # Value contains the value associated with the key in Name
            assert transformed_param["Value"] == params_dict[transformed_param["Name"]]

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_stop_pipeline_returns_status(self, mock_conn):
        mock_conn().describe_pipeline_execution.return_value = {"PipelineExecutionStatus": "Stopping"}

        hook = SageMakerHook(aws_conn_id="aws_default")
        pipeline_status = hook.stop_pipeline(pipeline_exec_arn="test")

        assert pipeline_status == "Stopping"
        mock_conn().stop_pipeline_execution.assert_called_once_with(PipelineExecutionArn="test")

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_stop_pipeline_raises_when_already_stopped_if_specified(self, mock_conn):
        error = ClientError(
            error_response={
                "Error": {"Message": "Only pipelines with 'Executing' status can be stopped", "Code": "0"}
            },
            operation_name="empty",
        )
        mock_conn().stop_pipeline_execution.side_effect = error
        mock_conn().describe_pipeline_execution.return_value = {"PipelineExecutionStatus": "Stopping"}

        hook = SageMakerHook(aws_conn_id="aws_default")
        with pytest.raises(ClientError) as raised_exception:
            hook.stop_pipeline(pipeline_exec_arn="test", fail_if_not_running=True)

        assert raised_exception.value == error

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_stop_pipeline_retries_on_conflict(self, mock_conn):
        conflict_error = ClientError(
            error_response={"Error": {"Code": "ConflictException"}},
            operation_name="empty",
        )
        mock_conn().stop_pipeline_execution.side_effect = [
            conflict_error,
            conflict_error,
            conflict_error,
            conflict_error,
            None,
        ]

        hook = SageMakerHook(aws_conn_id="aws_default")
        hook.stop_pipeline(pipeline_exec_arn="test")

        assert mock_conn().stop_pipeline_execution.call_count == 5

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_stop_pipeline_fails_if_all_retries_error(self, mock_conn):
        conflict_error = ClientError(
            error_response={"Error": {"Message": "blah", "Code": "ConflictException"}},
            operation_name="empty",
        )
        mock_conn().stop_pipeline_execution.side_effect = conflict_error

        hook = SageMakerHook(aws_conn_id="aws_default")
        with pytest.raises(ClientError) as raised_exception:
            hook.stop_pipeline(pipeline_exec_arn="test")

        assert mock_conn().stop_pipeline_execution.call_count == 5
        assert raised_exception.value == conflict_error

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_create_model_package_group(self, mock_conn):
        created = SageMakerHook().create_model_package_group("group-name")

        mock_conn().create_model_package_group.assert_called_once_with(
            ModelPackageGroupName="group-name",
            ModelPackageGroupDescription="",
        )
        assert created

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_create_model_package_group_returns_false_if_exists(self, mock_conn):
        mock_conn().create_model_package_group.side_effect = ClientError(
            error_response={
                "Error": {
                    "Code": "ValidationException",
                    "Message": "Model Package Group already exists: arn:aws:sagemaker:foo:bar",
                }
            },
            operation_name="empty",
        )
        hook = SageMakerHook()

        created = hook.create_model_package_group("group-name")

        assert created is False

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_create_auto_ml_parameter_structure(self, conn_mock):
        hook = SageMakerHook()

        hook.create_auto_ml_job(
            job_name="a",
            s3_input="b",
            target_attribute="c",
            s3_output="d",
            role_arn="e",
            compressed_input=True,
            time_limit=30,
            wait_for_completion=False,
        )

        assert conn_mock().create_auto_ml_job.call_args.kwargs == {
            "AutoMLJobConfig": {"CompletionCriteria": {"MaxAutoMLJobRuntimeInSeconds": 30}},
            "AutoMLJobName": "a",
            "InputDataConfig": [
                {
                    "CompressionType": "Gzip",
                    "DataSource": {"S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "b"}},
                    "TargetAttributeName": "c",
                }
            ],
            "OutputDataConfig": {"S3OutputPath": "d"},
            "RoleArn": "e",
        }

    @patch("airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook.conn", new_callable=mock.PropertyMock)
    def test_create_auto_ml_waits_for_completion(self, conn_mock):
        hook = SageMakerHook()
        conn_mock().describe_auto_ml_job.side_effect = [
            {"AutoMLJobStatus": "InProgress", "AutoMLJobSecondaryStatus": "a"},
            {"AutoMLJobStatus": "InProgress", "AutoMLJobSecondaryStatus": "b"},
            {
                "AutoMLJobStatus": "Completed",
                "AutoMLJobSecondaryStatus": "c",
                "BestCandidate": {"name": "me"},
            },
        ]

        ret = hook.create_auto_ml_job("a", "b", "c", "d", "e", check_interval=0)

        assert conn_mock().describe_auto_ml_job.call_count == 3
        assert ret == {"name": "me"}
